#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta
import os
import itertools

from matplotlib.dates import DateFormatter
import matplotlib.pyplot as plt
import numpy as np
import pandas
import scipy.stats
import scipy.ndimage as ndi
import scipy.optimize as opt
import statsmodels.api as sm

import doas

from smartcarb import io
from smartcarb import misc

import dplume

# kg/s -> Mt/y (Tg/y)
kgs_to_Mty = 1e3 * (365 * 24 * 60 * 60) / 1e12



def compute_emissions(nc):

    cats = [
       'CO2_point_snap1_major', # this was counted twice in Jan and Jul
       'CO2_point_snap1_minor',
       'CO2_point_snap2_minor',
       'CO2_point_snap3_minor',
       'CO2_point_snap4_minor',
       'CO2_point_snap5_minor',
       'CO2_point_snap34_minor',
       'CO2_point_snap8_minor',
       'CO2_point_snap9_minor',
       'CO2_area_construction',
       'CO2_line_aircraft',
       'CO2_point_aircraft',
       'CO2_area_aircraft',
       'CO2_area_forestry',
       'CO2_area_offroad',
       'CO2_line_shipping',
       'CO2_point_shipping',
       'CO2_line_train',
       'CO2_area_heating',
       'CO2_line_majorroads',
       'CO2_area_minorroads'
    ]

    total = 0.0
    for cat in cats:
        total += np.sum(nc.variables[cat][:] * 1235941.1)

    return total



def get_constant_emissions(time, use_constant=True):
    if not use_constant and time.year == 2015 and time.month in [1,7]:
        return 754.86
    else:
        return 534.30



def simple_leastsq(filename, error_std=0.7, seed='orbit',
        background_mode='model', use_constant_emissions=False, tracer='CO2'
    ):
    """\
    filename is either a file with 2d COSMO field or Level-2 data

    y[i,j] = x * H[i,j] + b

    - y: XCO2 measurements in plume
    - H: XCO2 from Berlin tracer (e.g. XCO2_BC) normalized by emissions
    - b: background (e.g. mean of "background" tracers in COSMO)

    The plume is defined where 'plume_tracer' > 'plume_thr'.

    If `use_const_emissions` use XCO2_BV in observation vector y.

    Names for H:
    """
    # tracer fields (synthetic observations)
    time = io.filename2time(filename)

    if tracer == 'CO2':
        co2_noise = error_std
        no2_noise = 'high'
    else:
        co2_noise = None
        no2_noise = error_std

    data = io.read_level2(filename, co2_noise=co2_noise, co2_thr=0.01,
                          no2_noise=no2_noise, no2_thr=0.30,
                          use_constant=use_constant_emissions,
                          add_berlin=True)

    # signals (constant and variable)
    if tracer == 'CO2':
        bc = io.read_fields(filename, ['XCO2_BC'], time,
                            correct_berlin=True)[0]
    else:
        bc = io.read_fields(filename, ['NO2_BC'], time, correct_berlin=True)[0]


    # define plume based on signal and cloud threshold
    if tracer == 'CO2':
        plume = dplume.create_true_plume(data, 'CO2', 'Berlin',
                                         signal_thr=0.05, cloud_thr=1.0)
        plume = np.array(plume) & np.array(data.clouds <= 0.01)
    else:
        plume = dplume.create_true_plume(data, 'NO2', 'Berlin',
                                         signal_thr=0.2e15, cloud_thr=1.0)
        plume = np.array(plume) & np.array(data.clouds <= 0.30)

    # remove data gaps (high uncertainty)
    plume &= np.array(np.isfinite(data[tracer]) & np.isfinite(data.clouds))


    if np.sum(plume) <= 1:
        return np.nan, np.nan, np.sum(plume)


    # background
    b = np.array(data[tracer] - data[tracer+'_noise'] -
                 data[tracer+'_true_signal'])[plume]

    if background_mode == 'model':
        bg_std = 0.0
    elif background_mode == 'mean':
        bg_std = b.std()
        b = b.mean()
    else:
        raise ValueError('unknown mode for background estimation')

    # observation vector
    y = np.array(data[tracer])[plume]
    #y = np.array(data[tracer] - data[tracer+'_noise'])[plume]

    # normalized footprint using constant emissions from simulations
    if tracer == 'CO2':
        constant_emissions = 534.2953988399561 # kg/s
    else:
        constant_emissions = 0.5868244173939502 # kg/s

    H = bc[plume] / constant_emissions

    # Se, xa and Sa
    if error_std == 0.0:
        Se = np.nan
    else:
        if tracer == 'CO2':
            Se = np.full(y.shape, error_std**2)
        else:
            no2_noise = {'low': 1e15, 'high': 2e15, 'S5': 1.3e15}[error_std]
            Se = np.full(y.shape, no2_noise**2)

    # forward model
    f = lambda x,_: H * x / kgs_to_Mty + b

    x0 = np.array([constant_emissions]) * kgs_to_Mty
    x, res = doas.solver.gauss_newton(x0, y, f, Se=Se)

    if not res['success']:
        print('fit failed')
        return np.nan, np.nan, np.sum(plume)

    return x[0], np.sqrt(float(res['S'])), np.sum(plume)



def main(out_filename, error_std=0.7, nmin=0, use_constant_emissions=False,
         background_mode='model', tracer='CO2', level2_filenames=None):

    if level2_filenames is None:
        filename = "/project/kug/smartcarb/results/Berlin/BerlinPlume.dat"
        data = pandas.read_csv(filename, sep='\t', index_col=2, parse_dates=True)

        level2_filenames = [
            io.create_filename(data.index[i], None, 'level2', 'Sentinel_7_CO2',
                               data.orbit[i], data.lon_eq[i])
            for i in range(data.shape[0])
            if data['pixels (<1%)'][i] > nmin
        ]

    # true emissions
    em_filename = '/output/SMARTCARB/output/timeseries/ts_%s_emissions.dat' % tracer
    em = pandas.read_csv(em_filename, delimiter='\t', index_col=0, parse_dates=True)
    em = em['Berlin (Mt/yr)']

    with open(out_filename, 'w') as out_file:

        if tracer == 'CO2':
            out_file.write('\t'.join([
                'UTC',
                'lon_eq',
                'orbit',
                'number of pixels',
                'CO2 emissions (Mt/yr)',
                'CO2 uncertainty (Mt/yr)',
                'CO2 true emissions (Mt/yr)',
            ]))
        else:
            out_file.write('\t'.join([
                'UTC',
                'lon_eq',
                'orbit',
                'number of pixels',
                'NO2 emissions (kt/yr)',
                'NO2 uncertainty (kt/yr)',
                'NO2 true emissions (kt/yr)',
            ]))

        out_file.write('\n')

        for fn in level2_filenames:

            _, orbit, time, lon_eq = io.process_filename(fn)

            x, x_std, N = simple_leastsq(
                fn, error_std=error_std, tracer=tracer,
                use_constant_emissions=use_constant_emissions,
                background_mode=background_mode
            )
            if N > 0:

                if use_constant_emissions:
                    true_em = {
                        'NO2': 18.506095e-3, # Mt/yr
                        'CO2': 16.849540
                    }[tracer]
                else:
                    true_em = em[time]

                if tracer == 'CO2':
                    print(time, N, x, true_em)
                else:
                    print(time, N, 1e3*x, 1e3*true_em)

                out_file.write('\t'.join([
                    str(time),
                    str(lon_eq),
                    str(orbit),
                    str(N),
                    '%10.3f' % (x if tracer == 'CO2' else 1e3 * x),
                    '%10.3f' % (x_std if tracer == 'CO2' else 1e3 * x_std),
                    '%10.3f' % (true_em if tracer == 'CO2' else 1e3 * true_em)
                ]))
                out_file.write('\n')





def read_results(filename, nsat=1):
    ts = pandas.read_csv(filename, delimiter='\t', index_col=0,
                         parse_dates=True, na_values='%10.3f' % np.nan)
    ts = misc.filter_dataset(ts, nsat)

    return ts


def plot_scatter(filename, nsat=1):
    ts = read_results(filename, nsat)

    est = ts['CO2 emissions (Mt/yr)']
    true = ts['CO2 true emissions (Mt/yr)']

    fig = plt.figure(figsize=(5,5))
    ax = fig.add_subplot(111, aspect='equal')

    ax.plot(true, est, 'o')

    ax.set_xlim(xmin=0)
    ax.set_ylim(ymin=0)


def sqrt_regression(n,y):

    x = 1.0 / np.sqrt(n)

    def model(c, b=None):
        return np.polyval(c,x)

    c0 = np.array([1.0, 0.0])
    c, res = doas.solver.gauss_newton(c0, y, model)
    c_std = np.sqrt(res['S'].diagonal())

    return c, c_std



def plot_vs_number(filename, nsat):

    fig = plt.figure()
    ax = fig.add_subplot(111)

    ts = read_results(filename, nsat)

    est = ts['CO2 emissions (Mt/yr)']
    true = ts['CO2 true emissions (Mt/yr)']

    n = ts['number of pixels']

    valids = n > 5
    n = n[valids]
    y = ts['CO2 uncertainty (Mt/yr)'][valids]

    x = 1.0 / np.sqrt(n)
    c, c_std = sqrt_regression(n, y)


    n_fit = np.arange(0, n.max())
    x_fit = 1.0 / np.sqrt(n_fit)

    y_fit = np.polyval(c, x_fit)
    u_fit = np.sqrt(c_std[0]**2 * x_fit**2 + c_std[1]**2)


    for n_pixels in [100,300,600]:
        v =  np.polyval(c, 1.0 / np.sqrt(n_pixels))
        print(n_pixels, '%.2f Mt/yr (%.0f%%)' % (v, 100*v/16.85))

    ax.fill_between(n_fit, y_fit - 3.0 * u_fit, y_fit + 3.0 * u_fit)
    ax.plot(n_fit, y_fit, 'k-', label='fit ($\pm$3$\sigma$)')
    ax.plot(n, y, 'o', ms=3, label='data', color='#ff7f0e')

    # compute number of pixels required for a given precision
    line = []
    for thr in [10,25,50,100]: # precision (in %)
    #for thr in [5,10]: # precision (in %)
        thr = thr / 100 * 16.85
        v = (thr - c[1]) / c[0]

        v_std = (-(thr - c[1]) / c[0]**2)**2 * c_std[0]**2
        v_std += (-1.0 / c[0])**2 * c_std[1]**2
        v_std = np.sqrt(v_std)

        vmin = 1.0 / (v - v_std)**2
        vmax = 1.0 / (v + v_std)**2
        vmed = 1.0 / v**2

        print('%d(+%d,-%d)' % (round(vmed), round(abs(vmed-vmin)),
                               round(abs(vmed-vmax))))

        line.append('%d-%d (%d)' % (round(vmin), round(vmax), round(vmed)))

        #ax.plot([vmin,vmax], [thr,thr], 'ro-')

    #print('\t'.join(line))


    ax.set_xlabel('number of pixels')
    ax.set_ylabel('CO$_2$ uncertainty (Mt/yr)')

    ax.set_xlim(0,1500)
    ax.set_ylim(0,50)

    ax.grid(True)

    plt.legend()
    plt.tight_layout()



def calculate_scatter(filename, nsat=1, nmin=0, nmax=np.inf):
    """
    """
    ts = read_results(filename, nsat)

    n = ts['number of pixels']
    valids = (nmin <= n) & (n < nmax)

    est = ts['CO2 emissions (Mt/yr)'][valids]
    unc = ts['CO2 uncertainty (Mt/yr)'][valids]

    true = ts['CO2 true emissions (Mt/yr)'][valids]

    if sum(valids) >= 10:
        scatter = np.std(est - true)
    else:
        scatter = np.nan

    return np.median(unc), scatter, 100 * scatter / true.mean(), sum(valids)



def rmsd_for_nsat(filename, nsat=1, use_constant=False, origin='Berlin',
                  is_mce=True, nmax=None):
    em10, em11, em_mean = read_true_emissions(use_constant, origin)

    ts6 = read_results(filename, nsat=None)
    #lon_eqs = sorted([int(v) for v in set(ts6.lon_eq)])
    lon_eqs = [0, 401, 805, 1207, 1610, 2012]

    n_plumes = []
    rmsd = []

    for lons in itertools.combinations(lon_eqs, nsat):

        mask = np.array([int(v) in lons for v in ts6.lon_eq], bool)
        ts = ts6[mask]

        if np.any(mask):
            est, _, _ = filter_results(ts, is_mce=is_mce, nmax=nmax)
            d = est - em_mean.loc[est.index]

            n = len(d)
            mb = np.mean(d)
            std = np.std(d)
            rms = np.sqrt(mb**2 + std**2 / n)

        else:
            n = 0
            rms = np.nan

        n_plumes.append(n)
        rmsd.append(rms)


    return np.ma.array(n_plumes), np.ma.array(rmsd, mask=np.isnan(rmsd))


def plot_rms(rms, ax=None, label=''):

    if ax is None:
        ax = plt.gca()

    nsats = np.arange(1,7)
    mini = [np.nanmin(v) for v in rms]
    mean = [np.nanmean(v) for v in rms]
    maxi = [np.nanmax(v) for v in rms]

    lines = ax.plot(nsats, mean, 'o--', label=label, mec='k')
    ax.fill_between(nsats, mini, maxi, color=lines[0].get_color(), alpha=0.5)

    ax.set_xlim(0.95,6.05)
    ax.set_ylim(0,20)
    ax.grid(True)
    ax.set_xlabel('number of satellites')
    ax.set_ylabel('RMSD (Mt yr$^{-1}$)')




def boxplot(filenames, nsat=6, nmax=None, sample='all', indx=None,
            origin='Berlin', use_constant=False, vmin=-20, vmax=20,
            is_mce=False):
    """
    """
    em10, em11, em_mean = read_true_emissions(use_constant, origin)

    data = []
    unc = []
    for filename in filenames:
        ts = read_results(filename, nsat)

        est, u, _ = filter_results(ts, is_mce=is_mce, nmax=nmax)
        d = est - em_mean.loc[est.index]

        data.append(d)
        unc.append(u)


    if sample == 'same':
        ind = sorted(set(data[0].index) & set(data[1].index) & set(data[2].index))
        nn = [len(v) for v in data]
        data = [d[ind] for d in data]
    else:
        nn = None

    if indx is not None:
        data = [d[indx] for d in data]

    sep = ' & '
    line = sep.join(['noise', '#plumes', 'MB', 'MB', 'SD', 'SD',
                      'RMSD', 'RMSD', 'unc', 'unc'])
    print(line)

    for noise, d,u in zip([0.5, 0.7, 1.0], data, unc):

        # annual mean emissions at overpass used for computing relative errors
        if origin == 'Berlin':
            t = 100 / 20.0 # em_mean.mean()
        elif origin == 'Janschwalde':
            t = 100 / 40.0 # Janschwalde (at overpass)
        else:
            raise ValueError

        n = len(d)
        mb  = np.mean(d)
        std = np.std(d)
        rms = np.sqrt(mb**2 + std**2 / n)

        print('%.1f' % noise, end=sep)
        print('%5d' % n, end=sep)
        print('%5.1f%s%5.1f' % (mb, sep, mb*t), end=sep)
        print('%5.1f%s%5.1f' % (std, sep, std*t), end=sep)
        print('%5.2f%s%5.1f' % (rms, sep, rms*t), end=sep)
        print('%5.1f%s%5.1f' % (np.mean(u), sep, np.mean(u)*t))

    print('Mean emissions: %.2f Mt/yr' % 20.0) #em_mean.mean())

    fig = plt.figure(figsize=(3.3,3.5))
    ax = fig.add_subplot(111)

    for i,n in enumerate(data):
        if nn is None:
            ax.text(i+1,vmax+0.1, str(len(n)), ha='center', va='bottom')
        else:
            text = '%s/%s' % (len(n), nn[i])
            ax.text(i+1,vmax+0.1, text, ha='center', va='bottom')


    ax.boxplot(data, showfliers=True, whis=[5,95], bootstrap=5000, widths=0.4,
               labels=['0.5', '0.7', '1.0'], meanline=True, showmeans=True)

    ax.set_xlabel('$\sigma_{VEG50}$ (ppm)')
    ax.set_ylabel('estimated $-$ true emissions (Mt/yr)')

    ax.set_ylim(vmin,vmax)
    ax.grid(True, ls='-', color='gray')
    plt.tight_layout(pad=0.25)
    fig.subplots_adjust(top=0.95)

    return data


def read_true_emissions(use_constant=False, origin='Berlin', tracer='CO2'):
    em10 = io.read_emissions(origin, 'overpass', use_constant=use_constant,
                            overpass=10, tracer=tracer)[0]
    em11 = io.read_emissions(origin, 'overpass', use_constant=use_constant,
                            overpass=11, tracer=tracer)[0]

    em_mean = pandas.Series(0.5 * (em10.values + em11.values), index=em11.index)

    return em10, em11, em_mean


def filter_results(ts, is_mce=False, nmax=None):
    """
    Remove bad fits of flux inversion.
    """

    # analytical inversion
    if 'CO2 emissions (Mt/yr)' in ts or 'NO2 emissions (kt/yr)' in ts:

        tracer = 'CO2' if 'CO2 emissions (Mt/yr)' in ts else 'NO2'

        if tracer == 'CO2':
            good = np.isfinite(ts['CO2 emissions (Mt/yr)'])
            good &= ts['CO2 uncertainty (Mt/yr)'] < 10.0

            return (
                ts['CO2 emissions (Mt/yr)'][good],
                ts['CO2 uncertainty (Mt/yr)'][good],
                ts.lon_eq[good]
            )
        else:
            good = np.isfinite(ts['NO2 emissions (kt/yr)'])
            good &= ts['NO2 uncertainty (kt/yr)'] < 10.0

            return (
                ts['NO2 emissions (kt/yr)'][good],
                ts['NO2 uncertainty (kt/yr)'][good],
                ts.lon_eq[good]
            )

    else: # mass-balance approach

        if is_mce:
            ts, ts_std, ts_sizes = dplume.io.average_results(ts, add_size=True)

        good = np.isfinite(ts['est. CO2 emissions (Mt/yr)'])

        if nmax is not None:
            if 'plume size' in ts:
                good &= ts['plume size'] > nmax
            else:
                print('Warning: number of pixels nmax not used.')

        if 'failed' in ts:
            good &= ts.failed == 0


        ts = ts[good]
        est = ts['est. CO2 emissions (Mt/yr)']
        lon_eqs = ts.lon_eq

        if is_mce:
            est_std = ts_std['est. CO2 emissions (Mt/yr)']
        else:
            est_std = ts['est. CO2 emissions uncertainty (Mt/yr)']

        return est, est_std, lon_eqs




def plot_ts(filename, nsat=1, use_constant=False, label=None,
           origin='Berlin', nmax=None, is_mce=False, tracer='CO2'):
    """
    Plot time series of estimated emissions.
    """

    ts = read_results(filename, nsat)
    lon_eqs_in_constellation = misc.lon_eq_by_nsat(nsat)

    em10, em11, em_mean = read_true_emissions(use_constant, origin,
                                              tracer=tracer)
    true = em_mean[ts.index]

    # remove bad values
    est, est_std, lon_eqs = filter_results(ts, is_mce=is_mce, nmax=nmax)
    true = true.loc[est.index]

    if tracer == 'NO2':
        em_mean = 1e3 * em_mean

    if origin == 'Berlin':
        #print(true.mean(), em_mean.mean())
        t = 100 / em_mean.mean()

    diff = est - true
    MB = np.mean(diff)
    SD = np.std(diff)
    RMSD = np.sqrt(MB**2 + SD**2 / len(diff))

    fig = plt.figure(figsize=(8.4, 3.5))
    ax = fig.add_subplot(111)

    ax.plot(em_mean.index, em_mean, 'k-',
            label='true emissions (10-11 UTC)')
    ax.fill_between(em10.index, em10, em11, color='silver')

    for i, lon_eq in enumerate(lon_eqs_in_constellation, 1):
        sat = lon_eqs == lon_eq

        values = est.loc[sat]
        errors = est_std.loc[sat]

        ax.errorbar(values.index, values, errors, ls='', marker='o',
                    label='#%d (%d plumes)' % (i, sum(sat)))

    if label is not None:
        for i,(t,v) in enumerate(zip(times,values)):
            s = '%.2g' % ts[label][i]
            ax.text(t,v,s)

    ax.xaxis.set_major_formatter(DateFormatter('%b'))

    ax.set_xlim('2015-01-01', '2015-12-31')
    if origin == 'Berlin':
        ax.set_ylim(-10,60)
    else:
        ax.set_ylim(ymin=0)

    ax.set_xlabel('2015')

    if tracer == 'CO2':
        ax.set_ylabel('CO$_2$ emissions (Mt/yr)')
    else:
        ax.set_ylabel('NO$_2$ emissions (kt/yr)')

    ax.grid(True, ls='-', color='gray')

    plt.legend(loc=1, ncol=3)
    plt.tight_layout(pad=0.25)
    fig.subplots_adjust(top=0.95)

    return ts



