


"""
Estimate time profiles of emissions using satellite observations.


annual time profile
- low-order polynomial
- periodic boundary conditions
- a priori state vector and covariance matrix
- ...

weekly cycle:
- weekday and weekend value
"""

import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate
import scipy.optimize

import doas
import smartcarb.io


def read_emissions():
    co2 = smartcarb.io.read_emissions('Berlin', 'raw', use_constant=False,
                                      tracer='CO2')

    no2 = smartcarb.io.read_emissions('Berlin', 'raw', use_constant=False,
                                      tracer='NO2')
    return co2, no2


def plot_diurnal_cycle():
    """
    Plot diurnal cycle of CO2 and NO2 emissions as well as CO2:NOX emission
    ratios.
    """
    plt.figure()

    co2, no2 = read_emissions()

    hod = np.arange(0,24)
    co2 = np.array([np.mean(co2[co2.index.hour == h]) for h in hod])
    no2 = np.array([np.mean(no2[no2.index.hour == h]) for h in hod])

    lines = []
    lines += plt.plot(hod, co2, 'o-',
                      label='CO$_2$ emissions (Mt CO$_2$ yr$^{-1}$)')
    lines += plt.plot(hod, 1e3*no2, 's-',
                      label='NO$_\mathrm{x}$ emissions (kt NO$_2$ yr$^{-1}$)')

    plt.grid(True)
    plt.xlabel('hour of day (UTC)')
    plt.ylabel('emissions')

    plt.xticks(np.arange(0,25,3))
    plt.ylim(0,40)


    plt.twinx()
    lines += plt.plot(hod, co2/no2, 'rv-', label='CO$_2$:NO$_\mathrm{x}$ emission ratios')
    plt.ylabel('CO$_2$:NO$_\mathrm{x}$ emission ratios', color='red')

    plt.ylim(600,1400)
    plt.yticks(np.arange(600,1401,100), color='red')

    plt.legend(lines, [l.get_label() for l in lines])
    plt.tight_layout()



def plot_seasonal_cycle():
    plt.figure()

    co2, no2 = read_emissions()

    moy = np.arange(1,13)
    co2 = np.array([np.mean(co2[co2.index.month == m]) for m in moy])
    no2 = np.array([np.mean(no2[no2.index.month == m]) for m in moy])

    lines = []
    lines += plt.plot(moy, co2, 'o-',
                      label='CO$_2$ emissions (Mt CO$_2$ yr$^{-1}$)')
    lines += plt.plot(moy, 1e3*no2, 's-',
                      label='NO$_\mathrm{x}$ emissions (kt NO$_2$ yr$^{-1}$)')

    plt.grid(True)
    plt.xlabel('month of year')
    plt.ylabel('emissions')

    plt.xticks(np.arange(1,13), ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                                 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'])
    plt.ylim(0,40)


    plt.twinx()
    lines += plt.plot(moy, co2/no2, 'rv-', label='CO$_2$:NO$_\mathrm{x}$ emission ratios')
    plt.ylabel('CO$_2$:NO$_\mathrm{x}$ emission ratios', color='red')

    plt.ylim(600,1400)
    plt.yticks(np.arange(600,1401,100), color='red')

    plt.legend(lines, [l.get_label() for l in lines])
    plt.tight_layout()


def plot_weekly_cycle():
    plt.figure()

    co2, no2 = read_emissions()

    dow = np.arange(0,7)
    co2 = np.array([np.mean(co2[co2.index.weekday == d]) for d in dow])
    no2 = np.array([np.mean(no2[no2.index.weekday == d]) for d in dow])

    lines = []
    lines += plt.plot(dow, co2, 'o-',
                      label='CO$_2$ emissions (Mt CO$_2$ yr$^{-1}$)')
    lines += plt.plot(dow, 1e3*no2, 's-',
                      label='NO$_\mathrm{x}$ emissions (kt NO$_2$ yr$^{-1}$)')

    plt.grid(True)
    plt.xlabel('day of week')
    plt.ylabel('emissions')

    plt.xticks(np.arange(7), ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']),
    plt.ylim(0,40)


    plt.twinx()
    lines += plt.plot(dow, co2/no2, 'rv-', label='CO$_2$:NO$_\mathrm{x}$ emission ratios')
    plt.ylabel('CO$_2$:NO$_\mathrm{x}$ emission ratios', color='red')

    plt.ylim(600,1400)
    plt.yticks(np.arange(600,1401,100), color='red')

    plt.legend(lines, [l.get_label() for l in lines])
    plt.tight_layout()


    # compute slope


def hermite_spline(x, p, t):
    """
    Hermite spline with periodic boundary conditions.
    """
    delta = np.diff(p)
    m = np.zeros_like(p)
    m[0] = 0.5 * (delta[0] + delta[-1])
    m[1:-1] = 0.5 * (delta[1:] + delta[:-1])
    m[-1] = m[0]

    # compute spline
    y = np.zeros(x.shape, dtype='f8')
    k = 0
    for i in range(x.size):
        if x[i] > t[k+1] and k < t.size-2:
            k += 1

        s = (x[i] - t[k]) / (t[k+1] - t[k])

        y[i] += p[k] * (1 + 2*s) * (1 - s)**2
        y[i] += m[k] * s * (1 - s)**2
        y[i] += p[k+1] * s**2 * (3 - 2*s)
        y[i] += m[k+1] * s**2 * (s - 1)

    return y



class SeasonalCycle:
    def __init__(self, x, y, knots, ystd=1.0):
        self.x = x
        self.y = y
        self.ystd = ystd

        self.knots = knots
        self.w0 = np.full(knots.size-1, y.mean())


    def __call__(self, w, x=None):
        w = np.append(w, w[0])
        s = hermite_spline(self.x if x is None else x,
                           w, self.knots)

        return s


    def residual(self, w):
        res = (self(w) - self.y) / self.ystd
        return res



def fit(ts, ts_std, n_knots=4):

    # seconds per year
    knots = np.linspace(0, 31536000, n_knots+1)

    # x- and y-data for fit
    times = ts.index
    start = np.datetime64('2015-01-01', dtype='datetime64[s]')
    xdata =  (times - start).total_seconds().to_numpy()
    ydata = ts.values
    ystd  = ts_std.values

    # remove invalid data
    valids = np.isfinite(ydata) & (ystd < 10)

    xdata = xdata[valids]
    ydata = ydata[valids]
    ystd = ystd[valids]

    # model
    cycle = SeasonalCycle(xdata, ydata, knots, ystd=ystd)

    # fit
    res = scipy.optimize.least_squares(cycle.residual, cycle.w0, method='lm')
    x = res.x

    # hourly values
    seconds = np.arange(0, 31536001, 60*60)
    times = start + seconds.astype('timedelta64[s]')

    return res, cycle, times, cycle(x, seconds)


def main():
    pass
